#include "arm.h"
#include "safeg.h"
#include "entry.h"
#include "safeg_measures.h"

typedef void (*startfunc) (void);

extern unsigned long NtContextSavedArea, NtContextSavedAreaEnd,
  TContextSavedArea, TContextSavedAreaEnd;
extern void SafeGStart(unsigned long *, unsigned long *);
extern unsigned long MonitorVectorTableBase;

#define STACK_SIZE 128          /* 128 byte stack */
unsigned long smm_stack[STACK_SIZE / 4];

/* stack top 򥢥֥Τꤹ */
unsigned long *const _smm_stack = &smm_stack[STACK_SIZE / 4 - 1];

const unsigned long nt_start_address = normal_start_address;

#if SAFEG_MEASURE_PERF_ENA == 1
unsigned long int safeg_latency[MAX_COUNT];
unsigned int latency_counter;
#endif /* SAFEG_MEASURE_PERF_ENA */

/*
 * safegȤνԤ
 */
int
safeg_init()
{
    unsigned long *addr = 0;

    /*
     * Monitor Vector Table Base Register ؤΥɥ쥹Ͽ 
     */
    CP15_MVBAR_WRITE(&MonitorVectorTableBase);

    /*
     * Non-Secure Access Control Register
     */
    CP15_NSACR_WRITE(NSACR_SETTING);

    /*
     * context areaν.
     * safeg ưOSꤷƤHWꥻåȻͤˤʤ褦,
     * Ŭڤͤꤹ뤳. 
     */
    for (addr = &NtContextSavedArea; addr < &NtContextSavedAreaEnd; addr++) {
        *addr = 0;
    }

    for (addr = &TContextSavedArea; addr < &TContextSavedAreaEnd; addr++) {
        *addr = 0;
    }

    return 0;
}

/*
 * safeg_init, monitor_init ˴ޤޤʤ򵭽Ҥ
 */
int
misc_init()
{
#ifdef ENABLE_FIQ_MANAGEMENT
    *((unsigned long *) FIQ_COUNT) = 0;
#endif

#ifdef ENABLE_IRQ_MANAGEMENT
    *((unsigned long *) IRQ_COUNT) = 0;
#endif

#if SAFEG_MEASURE_GPIO_ENA == 1
    gpio_init();
#endif

#if SAFEG_MEASURE_PERF_ENA == 1
    latency_counter = 0;
    asm("mcr p15, 0, %0, c15, c12, 0"::"r"(0)); /* counter disabled */
#endif

    return 0;
}

int
arm_main()
{
    safeg_init();
    misc_init();
    SafeGStart((unsigned long *) normal_start_address,
               (unsigned long *) secure_start_address);
    return 0;
}

#ifdef ENABLE_FIQ_MANAGEMENT
/* fiq_management ΥƥѴؿ */
void fiq_test()
{
	*((unsigned long *) FIQ_COUNT) += 1;
}

/*
 * fiq ɤ򤳤δؿƤӽФ.
 * FIQ, IRQ ػ߾֤ǸƤӽФ.
 * ƤӽФΥå쥸¸ƤΤ, å쥸˲
 * Ƥ褤.
 */
void fiq_management()
{
    /*
     * fiq_test(); 
     */
}
#endif

#ifdef ENABLE_IRQ_MANAGEMENT
/* irq_management ΥƥѴؿ */
void irq_test()
{
    *((unsigned long *) IRQ_COUNT) += 1;
}

/*
 * irq ɤ򤳤δؿƤӽФ.
 * IRQ ػ߾֤ǸƤӽФ.
 * ƤӽФΥå쥸¸ƤΤ, å쥸˲
 * Ƥ褤.
 */
void irq_management()
{
    /*
     * irq_test(); 
     */
}
#endif
